"use strict";
//
// Copyright 2021 Vulcanize, Inc.
//
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (g && (g = 0, op[0] && (_ = 0)), _) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getMappingSlot = exports.getValueByType = exports.getStorageValue = exports.getStorageInfo = void 0;
var ethers_1 = require("ethers");
/**
 * Function to get storage information of variable from storage layout.
 * @param storageLayout
 * @param variableName
 */
var getStorageInfo = function (storageLayout, variableName) {
    var storage = storageLayout.storage, types = storageLayout.types;
    var targetState = storage.find(function (state) { return state.label === variableName; });
    // Throw if state variable could not be found in storage layout.
    if (!targetState) {
        throw new Error('Variable not present in storage layout.');
    }
    return __assign(__assign({}, targetState), { slot: ethers_1.utils.hexlify(ethers_1.BigNumber.from(targetState.slot)), types: types });
};
exports.getStorageInfo = getStorageInfo;
/**
 * Function to get the value from storage for a contract variable.
 * @param storageLayout
 * @param getStorageAt
 * @param blockHash
 * @param address
 * @param variableName
 */
/* eslint-disable @typescript-eslint/no-explicit-any */
var getStorageValue = function (storageLayout, getStorageAt, blockHash, address, variableName) {
    var mappingKeys = [];
    for (var _i = 5; _i < arguments.length; _i++) {
        mappingKeys[_i - 5] = arguments[_i];
    }
    return __awaiter(void 0, void 0, void 0, function () {
        var _a, slot, offset, type, types;
        return __generator(this, function (_b) {
            _a = (0, exports.getStorageInfo)(storageLayout, variableName), slot = _a.slot, offset = _a.offset, type = _a.type, types = _a.types;
            return [2 /*return*/, getDecodedValue(getStorageAt, blockHash, address, types, { slot: slot, offset: offset, type: type }, mappingKeys)];
        });
    });
};
exports.getStorageValue = getStorageValue;
/**
 * Get value according to type described by the label.
 * @param storageValue
 * @param typeLabel
 */
var getValueByType = function (storageValue, typeLabel) {
    // Parse value for boolean type.
    if (typeLabel === 'bool') {
        return !ethers_1.BigNumber.from(storageValue).isZero();
    }
    var _a = __read(typeLabel.match(/^enum|((u?)int([0-9]+))/) || [false], 4), isIntegerOrEnum = _a[0], isInteger = _a[1], isUnsigned = _a[2], integerSize = _a[3];
    // Parse value for uint/int type or enum type.
    if (isIntegerOrEnum) {
        var bigNumber = ethers_1.BigNumber.from(storageValue);
        if (Boolean(isInteger) && !isUnsigned) {
            bigNumber = bigNumber.fromTwos(Number(integerSize));
        }
        return BigInt(bigNumber.toString());
    }
    // Parse value for string type.
    if (typeLabel.includes('string')) {
        return ethers_1.utils.toUtf8String(storageValue);
    }
    if (typeLabel.startsWith('address')) {
        return ethers_1.utils.getAddress(storageValue);
    }
    return storageValue;
};
exports.getValueByType = getValueByType;
/**
 * Function to get decoded value according to type and encoding.
 * @param getStorageAt
 * @param blockHash
 * @param address
 * @param types
 * @param storageInfo
 */
/* eslint-disable @typescript-eslint/no-explicit-any */
var getDecodedValue = function (getStorageAt, blockHash, address, types, storageInfo, mappingKeys) { return __awaiter(void 0, void 0, void 0, function () {
    var slot, offset, type, _a, encoding, numberOfBytes, typeLabel, base, mappingValueType, mappingKeyType, members, value, proof, _b, isArray, arraySize, isStruct, _c, mappingSlot, _d, dynamicArraySlot, size;
    var _e, _f;
    return __generator(this, function (_g) {
        switch (_g.label) {
            case 0:
                slot = storageInfo.slot, offset = storageInfo.offset, type = storageInfo.type;
                _a = types[type], encoding = _a.encoding, numberOfBytes = _a.numberOfBytes, typeLabel = _a.label, base = _a.base, mappingValueType = _a.value, mappingKeyType = _a.key, members = _a.members;
                _b = __read(typeLabel.match(/\[([0-9]+)\]$/) || [false], 2), isArray = _b[0], arraySize = _b[1];
                // If variable is array type.
                if (Boolean(isArray) && base) {
                    return [2 /*return*/, getArrayValue(getStorageAt, blockHash, address, types, mappingKeys, slot, base, Number(arraySize))];
                }
                isStruct = /^struct .+/.test(typeLabel);
                // If variable is struct type.
                if (isStruct && members) {
                    return [2 /*return*/, getStructureValue(getStorageAt, blockHash, address, types, mappingKeys, slot, members)];
                }
                _c = encoding;
                switch (_c) {
                    case 'inplace': return [3 /*break*/, 1];
                    case 'bytes': return [3 /*break*/, 3];
                    case 'mapping': return [3 /*break*/, 5];
                    case 'dynamic_array': return [3 /*break*/, 6];
                }
                return [3 /*break*/, 10];
            case 1: return [4 /*yield*/, getInplaceValue(getStorageAt, blockHash, address, slot, offset, numberOfBytes)];
            case 2:
                (_e = _g.sent(), value = _e.value, proof = _e.proof);
                return [3 /*break*/, 11];
            case 3: return [4 /*yield*/, getBytesValue(getStorageAt, blockHash, address, slot)];
            case 4:
                (_f = _g.sent(), value = _f.value, proof = _f.proof);
                return [3 /*break*/, 11];
            case 5:
                {
                    if (mappingValueType && mappingKeyType) {
                        mappingSlot = (0, exports.getMappingSlot)(types, slot, mappingKeyType, mappingKeys[0]);
                        return [2 /*return*/, getDecodedValue(getStorageAt, blockHash, address, types, { slot: mappingSlot, offset: 0, type: mappingValueType }, mappingKeys.slice(1))];
                    }
                    else {
                        throw new Error("Mapping value type not specified for ".concat(mappingKeys[0]));
                    }
                    return [3 /*break*/, 11];
                }
                _g.label = 6;
            case 6:
                if (!base) return [3 /*break*/, 8];
                return [4 /*yield*/, getDynamicArrayInfo(getStorageAt, blockHash, address, slot, offset, numberOfBytes)];
            case 7:
                _d = _g.sent(), dynamicArraySlot = _d.slot, size = _d.size;
                return [2 /*return*/, getArrayValue(getStorageAt, blockHash, address, types, mappingKeys, dynamicArraySlot, base, size)];
            case 8: throw new Error('Missing base type for dynamic array.');
            case 9: return [3 /*break*/, 11];
            case 10: throw new Error("Encoding ".concat(encoding, " not implemented."));
            case 11: return [2 /*return*/, {
                    value: (0, exports.getValueByType)(value, typeLabel),
                    proof: proof
                }];
        }
    });
}); };
/**
 * Function to get slot for mapping types.
 * @param mappingSlot
 * @param key
 */
var getMappingSlot = function (types, mappingSlot, keyType, key) {
    var _a = types[keyType], encoding = _a.encoding, typeLabel = _a.label;
    // If key is boolean type convert to 1 or 0 which is the way value is stored in memory.
    if (typeLabel === 'bool') {
        key = key ? 1 : 0;
    }
    // If key is string convert to hex string representation.
    if (typeLabel === 'string' && typeof key === 'string') {
        key = ethers_1.utils.hexlify(ethers_1.utils.toUtf8Bytes(key));
    }
    // If key is still boolean type the argument passed as key is invalid.
    if (typeof key === 'boolean') {
        throw new Error('Invalid key.');
    }
    // https://github.com/ethers-io/ethers.js/issues/1079#issuecomment-703056242
    var mappingSlotPadded = ethers_1.utils.hexZeroPad(mappingSlot, 32);
    var keyPadded = encoding === 'bytes'
        ? ethers_1.utils.hexlify(key)
        : ethers_1.utils.hexZeroPad(ethers_1.utils.hexlify(key), 32);
    // https://docs.soliditylang.org/en/v0.8.4/internals/layout_in_storage.html#mappings-and-dynamic-arrays
    var fullKey = ethers_1.utils.concat([
        keyPadded,
        mappingSlotPadded
    ]);
    var slot = ethers_1.utils.keccak256(fullKey);
    return slot;
};
exports.getMappingSlot = getMappingSlot;
var getDynamicArrayInfo = function (getStorageAt, blockHash, address, slot, offset, numberOfBytes) { return __awaiter(void 0, void 0, void 0, function () {
    var value, size, paddedSlot;
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0: return [4 /*yield*/, getInplaceValue(getStorageAt, blockHash, address, slot, offset, numberOfBytes)];
            case 1:
                value = (_a.sent()).value;
                size = Number((0, exports.getValueByType)(value, 'uint'));
                paddedSlot = ethers_1.utils.hexZeroPad(slot, 32);
                slot = ethers_1.utils.keccak256(paddedSlot);
                return [2 /*return*/, { size: size, slot: slot }];
        }
    });
}); };
var getArrayValue = function (getStorageAt, blockHash, address, types, mappingKeys, slot, base, arraySize) { return __awaiter(void 0, void 0, void 0, function () {
    var resultArray, proofs, baseNumberOfBytes, getArrayElement, _a, arrayIndex, remainingKeys, i, _b, value, proof;
    return __generator(this, function (_c) {
        switch (_c.label) {
            case 0:
                resultArray = [];
                proofs = [];
                baseNumberOfBytes = types[base].numberOfBytes;
                getArrayElement = function (mappingKeys, index) { return __awaiter(void 0, void 0, void 0, function () {
                    var arraySlotOffset, slotIndex, elementsInSlot, slotsUsedByElement, arraySlot;
                    return __generator(this, function (_a) {
                        arraySlotOffset = 0;
                        if (Number(baseNumberOfBytes) <= 32) {
                            elementsInSlot = Math.floor(32 / Number(baseNumberOfBytes));
                            slotIndex = Math.floor(index / elementsInSlot);
                            arraySlotOffset = (index % elementsInSlot) * Number(baseNumberOfBytes);
                        }
                        else {
                            slotsUsedByElement = Math.ceil(Number(baseNumberOfBytes) / 32);
                            slotIndex = slotsUsedByElement * index;
                        }
                        arraySlot = ethers_1.BigNumber.from(slot).add(slotIndex).toHexString();
                        return [2 /*return*/, getDecodedValue(getStorageAt, blockHash, address, types, { slot: arraySlot, offset: arraySlotOffset, type: base }, mappingKeys)];
                    });
                }); };
                _a = __read(mappingKeys), arrayIndex = _a[0], remainingKeys = _a.slice(1);
                if (typeof arrayIndex === 'number') {
                    return [2 /*return*/, getArrayElement(remainingKeys, arrayIndex)];
                }
                i = 0;
                _c.label = 1;
            case 1:
                if (!(i < arraySize)) return [3 /*break*/, 4];
                return [4 /*yield*/, getArrayElement(mappingKeys, i)];
            case 2:
                _b = _c.sent(), value = _b.value, proof = _b.proof;
                resultArray.push(value);
                // Each element in array gets its own proof even if it is packed.
                proofs.push(JSON.parse(proof.data));
                _c.label = 3;
            case 3:
                i++;
                return [3 /*break*/, 1];
            case 4: return [2 /*return*/, {
                    value: resultArray,
                    proof: {
                        data: JSON.stringify(proofs)
                    }
                }];
        }
    });
}); };
var getStructureValue = function (getStorageAt, blockHash, address, types, mappingKeys, slot, members) { return __awaiter(void 0, void 0, void 0, function () {
    var getStructMember, _a, memberName, remainingKeys, member, resultPromises, results, initialValue;
    return __generator(this, function (_b) {
        switch (_b.label) {
            case 0:
                getStructMember = function (mappingKeys, member) { return __awaiter(void 0, void 0, void 0, function () {
                    var structSlot;
                    return __generator(this, function (_a) {
                        structSlot = ethers_1.BigNumber.from(slot).add(member.slot).toHexString();
                        return [2 /*return*/, getDecodedValue(getStorageAt, blockHash, address, types, { slot: structSlot, offset: member.offset, type: member.type }, mappingKeys)];
                    });
                }); };
                _a = __read(mappingKeys), memberName = _a[0], remainingKeys = _a.slice(1);
                member = members.find(function (member) { return member.label === memberName; });
                // If member name passed in argument is present.
                if (member) {
                    return [2 /*return*/, getStructMember(remainingKeys, member)];
                }
                resultPromises = members.map(function (member) { return __awaiter(void 0, void 0, void 0, function () {
                    return __generator(this, function (_a) {
                        return [2 /*return*/, getStructMember(mappingKeys, member)];
                    });
                }); });
                return [4 /*yield*/, Promise.all(resultPromises)];
            case 1:
                results = _b.sent();
                initialValue = {
                    value: {},
                    proof: { data: JSON.stringify({}) }
                };
                // Return struct type value as an object with keys as the struct member labels.
                return [2 /*return*/, members.reduce(function (acc, member, index) {
                        acc.value[member.label] = results[index].value;
                        var proofData = JSON.parse(acc.proof.data);
                        proofData[member.label] = JSON.parse(results[index].proof.data);
                        acc.proof.data = JSON.stringify(proofData);
                        return acc;
                    }, initialValue)];
        }
    });
}); };
/**
 * Function to get value for inplace encoding.
 * @param address
 * @param slot
 * @param offset
 * @param numberOfBytes
 * @param getStorageAt
 */
var getInplaceValue = function (getStorageAt, blockHash, address, slot, offset, numberOfBytes) { return __awaiter(void 0, void 0, void 0, function () {
    var _a, value, proof, valueLength, start, end;
    return __generator(this, function (_b) {
        switch (_b.label) {
            case 0: return [4 /*yield*/, getStorageAt({ blockHash: blockHash, contract: address, slot: slot })];
            case 1:
                _a = _b.sent(), value = _a.value, proof = _a.proof;
                valueLength = ethers_1.utils.hexDataLength(value);
                start = valueLength - (offset + Number(numberOfBytes));
                end = valueLength - offset;
                return [2 /*return*/, {
                        value: ethers_1.utils.hexDataSlice(value, start, end),
                        proof: proof
                    }];
        }
    });
}); };
/**
 * Function to get value for bytes encoding.
 * @param address
 * @param slot
 * @param getStorageAt
 */
var getBytesValue = function (getStorageAt, blockHash, address, slot) { return __awaiter(void 0, void 0, void 0, function () {
    var _a, value, proof, proofs, slotValue, length, lastByteHex, hexStringArray, paddedSlotHex, position, i, _b, value_1, proof_1;
    return __generator(this, function (_c) {
        switch (_c.label) {
            case 0: return [4 /*yield*/, getStorageAt({ blockHash: blockHash, contract: address, slot: slot })];
            case 1:
                _a = _c.sent(), value = _a.value, proof = _a.proof;
                proofs = [JSON.parse(proof.data)];
                slotValue = ethers_1.BigNumber.from(value);
                length = 0;
                // Get length of bytes stored.
                // https://docs.soliditylang.org/en/v0.7.6/internals/layout_in_storage.html#bytes-and-string
                // Check if the lowest bit is set.
                if ((slotValue.and(1)).toNumber() !== 0) {
                    // If the lowest bit is set, the value is an odd number.
                    // So subtract 1 and divide by 2 to get the length.
                    length = slotValue.sub(1).div(2).toNumber();
                }
                else {
                    lastByteHex = ethers_1.utils.hexDataSlice(value, 31, 32);
                    length = ethers_1.BigNumber.from(lastByteHex).div(2).toNumber();
                }
                // Get value from the byte array directly if length is less than 32.
                if (length < 32) {
                    return [2 /*return*/, {
                            value: ethers_1.utils.hexDataSlice(value, 0, length),
                            proof: {
                                data: JSON.stringify(proofs)
                            }
                        }];
                }
                hexStringArray = [];
                paddedSlotHex = ethers_1.utils.hexZeroPad(slot, 32);
                position = ethers_1.utils.keccak256(paddedSlotHex);
                i = 0;
                _c.label = 2;
            case 2:
                if (!(i < length / 32)) return [3 /*break*/, 5];
                return [4 /*yield*/, getStorageAt({
                        blockHash: blockHash,
                        contract: address,
                        slot: ethers_1.BigNumber.from(position).add(i).toHexString()
                    })];
            case 3:
                _b = _c.sent(), value_1 = _b.value, proof_1 = _b.proof;
                hexStringArray.push(value_1);
                proofs.push(JSON.parse(proof_1.data));
                _c.label = 4;
            case 4:
                i++;
                return [3 /*break*/, 2];
            case 5: 
            // Slice trailing bytes according to length of value.
            return [2 /*return*/, {
                    value: ethers_1.utils.hexDataSlice(ethers_1.utils.hexConcat(hexStringArray), 0, length),
                    proof: {
                        data: JSON.stringify(proofs)
                    }
                }];
        }
    });
}); };
//# sourceMappingURL=storage.js.map